@@ -1,6 +1,7 @@
 from bz2 import BZ2File
 from collections import Counter, Sequence, Iterable, \
     Mapping
+from functools import partial
 import gc
 from email.mime.text import MIMEText
 import inspect
@@ -905,7 +906,55 @@ def max_key(d, fn=identity):
     return max(d.items(), key=lambda x: fn(x[1]))[0]
 
 
-def fgrep(text, term, window=25):
+def is_builtin(x, drop_callables=True):
+    """Check if an object is a Python built-in object.
+
+    Parameters
+    ----------
+    x: object
+    drop_callables: bool
+        If True, return False for callables (basically functions, methods, or
+        classes). These typically will return True otherwise since they are of
+        class `type` or `builtin_function_or_method`.
+
+    Returns
+    -------
+    bool: True if `x` is a built-in object, False otherwise.
+    """
+    def _builtin(x, drop_callables):
+        if callable(x) and drop_callables:
+            return False
+        return x.__class__.__module__ == 'builtins'
+    builtin = partial(_builtin, drop_callables=drop_callables)
+    # Check mapping first because mappings are iterable.
+    if isinstance(x, Mapping):
+        return builtin(x) and all(builtin(o) for o in flatten(x.items()))
+    elif isinstance(x, Iterable):
+        return builtin(x) and all(builtin(o) for o in flatten(x))
+    return builtin(x)
+
+
+def hashable(x):
+    """Check if an object is hashable. Hashable objects will usually be
+    immutable though this is not guaranteed.
+
+    Parameters
+    ----------
+    x: object
+        The item to check for hashability.
+
+    Returns
+    -------
+    bool: True if `x` is hashable (suggesting immutability), False otherwise.
+    """
+    try:
+        _ = hash(x)
+        return True
+    except TypeError:
+        return False
+
+
+def fgrep(text, term, window=25, with_idx=False, reverse=False):
     """Search a string for a given term. If found, print it with some context.
     Similar to `grep -C 1 term text`. `fgrep` is short for faux grep.
 
@@ -917,16 +966,23 @@ def fgrep(text, term, window=25):
         Term to look for in text.
     window: int
         Number of characters to display before and after the matching term.
+    with_idx: bool
+        If True, return index as well as string.
+    reverse: bool
+        If True, reverse search direction (find last match rather than first).
 
     Returns
     -------
-    str: The desired term and its surrounding context. If the term isn't
-    present, an empty string is returned.
+    str or tuple[int, str]: The desired term and its surrounding context.
+        If the term isn't present, an empty string is returned. If
+        with_idx=True, a tuple of (match index, string with text) is returned.
     """
-    idx = text.find(term)
+    idx = text.rfind(term) if reverse else text.find(term)
     if idx == -1:
-        return ''
-    return text[max(idx-window, 0):idx+window]
+        res = ''
+    else:
+        res = text[max(idx-window, 0):idx+window]
+    return (idx, res) if with_idx else res
 
 
 def spacer(char='-', n_chars=79, newlines_before=1, newlines_after=1):
